/**********************************************************************
** This program is part of the kinetics library and is
**           copyright (C) 1995 Upinder S. Bhalla.
** It is made available under the terms of the
**           GNU Library General Public License. 
** See the file COPYRIGHT for the full notice.
**********************************************************************/

/*
 * Boilerplate-copied from ./pool.c, ideas from ./{pool,reac,ksolve}.c
 */

#include "kin_ext.h"
#include <math.h>
#include <float.h>
#define SIMPLE_SPRNG
#include "sprng.h"
#include "string.h"

#define POOL_BUFFER 0x04

#ifdef __STDC__
static int gsolve_do_reset (struct gsolve_type *gsolve);
static int check_msg_type (struct gsolve_type *gsolve, MsgIn * msg);
void pq_insert(double, int);
void pq_swap(int, int);
void pq_update(int, double);
void pq_update_aux(int);
int add_dep_list(int, int);
/*
 * static void do_oldmsgmode (struct gsolve_type *gsolve, double *A,
 * double *B, double *conserve, double *sumtotal); 
 */
#else
static int gsolve_do_reset ();
static int check_msg_type ();
void pq_insert();
void pq_swap();
void pq_update();
void pq_update_aux();
int add_dep_list();
/*
 * static void do_oldmsgmode(); 
 */
#endif

#if 0
static long forward_reac_num = 0;
static long backward_reac_num = 0;
#endif

#if 0
typedef struct node_type {
	int index;
	struct node_type *next;
} node;
#endif

double propensity_trans[1900]; /* For Gibson-Bruck, stored in this
       flattened array instead of gsolve->prop*{reac,enz}. Not quite
       elegant. We still use gsolve->propensity_total though */

unsigned short dep_graph_adj[1900][1901];    /* adjacency list representation of
       dependency graph. [i][1900] store (in succession) indices of
       dependent transitions. [i][1901] stores the number of these. */
typedef struct priority_queue_node_type {
  double time;
  int transition_index;        /* priority queue -> transition index */
} pq_node;
pq_node pq_nodes[1900];   /* array representation of priority queue */
int pq_size;
int transition_pq_index[1900]; /* transition -> priority queue index */

double propensity_trans_zero[1900]; 
double time_zero[1900]; 
double trans_time_zero[1900];

GsolveFunc (gsolve, action)
     struct gsolve_type *gsolve;
     Action *action;
{
  MsgIn *msg;
  int i, j, k;
  double **data = NULL;
  struct reac_type *reac = NULL;
  struct enz_type *enz = NULL;
  struct reac_type *reac_chosen = NULL;
  struct enz_type *enz_chosen = NULL;
  int reac_chosen_direction = 0;	/* 1 or -1*/
  int enz_chosen_direction = 0;	/* 1, 2 or 3 */

  int trans_chosen = 1900;

  double deviate_transition = 0.0;
  double transition_time = 0.0, transition_steps = 0.0;
  double prop_acc = 0.0;
  double dt;
  double orig_nenz;
  
  int *nsteps;
  int *stepcount;
  int *current_step;
  float *start_time;
  float *duration;
  double *simulation_time;

  int r_or_e_type, r_or_e_index, r_or_e_direction;
  double minimum_time = DBL_MAX;
  double time;

  int **se;
  
  double propensity_trans_old;

  if (debug > 1)
    {
      ActionHeader ("ReactionGsolve", gsolve, action);
    }
  
  /*
   * add check to run only if stoch_flag is set 
   */
  
  SELECT_ACTION (action)
    {
    case PROCESS:
      if (gsolve->reac_list_populated || gsolve->enz_list_populated)
	{
	  gsolvePeek (&nsteps, &stepcount, &current_step, &start_time,
		      &duration, &simulation_time);
	  
	  if ((gsolve->method == 1) || (gsolve->method == 2)) {
	    /* build up propensity_reac[][] and propensity_total */
		gsolve->propensity_total = 0;
	    
		if (gsolve->reac_list_populated)
	      for (i = 0; i < gsolve->reac_list->nelements; i++)
		{
		  reac = (struct reac_type *) gsolve->reac_list->element[i];
		  data = reac->msgdata;
		  
		  gsolve->propensity_reac[0][i] = reac->kf;
		  for (j = 0; j < reac->msgcount[0]; j++, data++)
		    {
		      gsolve->propensity_reac[0][i] *= **data;
		    }
		  /* Skipping multiplication by dt as this is a relative choice algorithm */
		  gsolve->propensity_reac[1][i] = reac->kb;
		  for (j = 0; j < reac->msgcount[1]; j++, data++)
		    {
		      gsolve->propensity_reac[1][i] *= **data;
		    }
		  gsolve->propensity_total +=
		    gsolve->propensity_reac[0][i] + gsolve->propensity_reac[1][i];
		}
	    /* build up propensity_enz[][] and propensity_total */
	    if (gsolve->enz_list_populated)
	      for (i = 0; i < gsolve->enz_list->nelements; i++)
		{
		  enz = (struct enz_type *) gsolve->enz_list->element[i];
		  data = enz->msgdata;
		  
		  for (j = 0; j < enz->msgcount[0]; j++, data++)
		    gsolve->propensity_enz[0][i] = **data;
		  if (enz->usecomplex)
		    gsolve->propensity_enz[0][i] -= enz->nComplex;
		  if (gsolve->propensity_enz[0][i] < 1)
		    gsolve->propensity_enz[0][i] = 0;
		  for (j = 0; j < enz->msgcount[1]; j++, data++)
		    gsolve->propensity_enz[0][i] *= **data;
		  gsolve->propensity_enz[0][i] *= enz->k1;
		  
		  gsolve->propensity_enz[1][i] = enz->k2 * enz->nComplex;
		  if (enz->nComplex < 1)
		    gsolve->propensity_enz[1][i] = 0;
		  
		  gsolve->propensity_enz[2][i] = enz->k3 * enz->nComplex;
		  if (enz->nComplex < 1)
		    gsolve->propensity_enz[2][i] = 0;
		  
		  gsolve->propensity_total +=
		    gsolve->propensity_enz[0][i] + gsolve->propensity_enz[1][i] +
		    gsolve->propensity_enz[2][i];
		  
		  enz->pA = 0;
		}
	    
	    /* if prop total is zero, skip out now.
	     * set step and time counts appropriately first. */
	    if (!(gsolve->propensity_total))
	      {
		*current_step = ((*start_time + *duration) / ClockValue (0)) - 1;
		*stepcount = *current_step;	/* aaaaaaargh! */
		*nsteps = 1;
		*simulation_time = (*start_time + *duration) - ClockValue (0);
		
		gsolvePoke (&nsteps, &stepcount, &current_step, &simulation_time);
		goto exitprocess;
	      }
	    
	    if (gsolve->method == 1)
	      {
		deviate_transition = sprng () * gsolve->propensity_total;
		/* scale deviate to compare against propensity coeffs */
		
		/*
		 * choose reaction or enzyme transition, inelegantly 
		 */
		for (i = 0; i < gsolve->reac_list->nelements; i++)
		  {
		    if (deviate_transition <=
			(prop_acc + gsolve->propensity_reac[0][i]))
		      {
			reac_chosen =
			  (struct reac_type *) gsolve->reac_list->element[i];
			reac_chosen_direction = 1;	/* forward */
			/*              printf("forward: %ld\n", ++forward_reac_num);  */
			break;
		      }
		    else
		      prop_acc += gsolve->propensity_reac[0][i];
		    
		    if (deviate_transition <=
			(prop_acc + gsolve->propensity_reac[1][i]))
		      {
			reac_chosen =
			  (struct reac_type *) gsolve->reac_list->element[i];
			reac_chosen_direction = -1;	/* backward */
			/*              printf("backward: %ld\n", ++backward_reac_num);  */
			break;
		      }
		    else
		      prop_acc += gsolve->propensity_reac[1][i];
		  }
		if (!reac_chosen)
		  for (i = 0; i < gsolve->enz_list->nelements; i++)
		    {
		      if (deviate_transition <=
			  (prop_acc + gsolve->propensity_enz[0][i]))
			{
			  enz_chosen =
			    (struct enz_type *) gsolve->enz_list->element[i];
			  enz_chosen_direction = 1;
			  break;
			}
		      else
			prop_acc += gsolve->propensity_enz[0][i];
		      
		      if (deviate_transition <=
			  (prop_acc + gsolve->propensity_enz[1][i]))
			{
			  enz_chosen =
			    (struct enz_type *) gsolve->enz_list->element[i];
			  enz_chosen_direction = 2;
			  break;
			}
		      else
			prop_acc += gsolve->propensity_enz[1][i];
		      
		      if (deviate_transition <=
			  (prop_acc + gsolve->propensity_enz[2][i]))
			{
			  enz_chosen =
			    (struct enz_type *) gsolve->enz_list->element[i];
			  enz_chosen_direction = 3;
			  break;
			}
		      else
			prop_acc += gsolve->propensity_enz[2][i];
		      
		    }		/* CAUTION: replicated code! */
		
		/*
		 * choose time and update current time 
		 */
		
		transition_time = -1 * log (sprng ()) / gsolve->propensity_total;
	      }
	    else if (gsolve->method == 2)
	      {			/* Gillespie's First Reaction Method */
		
		for (i = 0; i < gsolve->reac_list->nelements; i++)
		  {
		    if (gsolve->propensity_reac[0][i]) {
		      time = -1 * log (sprng ()) / gsolve->propensity_reac[0][i];
		      if (time < minimum_time)
			{
			  minimum_time = time;
			  r_or_e_type = 1;	/* reac chosen */
			  r_or_e_index = i;
			  r_or_e_direction = 1;	/* forward */
			}
		    } else
		      time = DBL_MAX;

		    if (gsolve->propensity_reac[1][i]) {
		      time = -1 * log (sprng ()) / gsolve->propensity_reac[1][i];
		      if (time < minimum_time)
			{
			  minimum_time = time;
			  r_or_e_type = 1;	/* reac chosen */
			  r_or_e_index = i;
			  r_or_e_direction = -1;	/* backward */
			}
		    } else
		      time = DBL_MAX;
		  }
		
		for (i = 0; i < gsolve->enz_list->nelements; i++)
		  {
		    if (gsolve->propensity_enz[0][i]) {
		      time = -1 * log (sprng ()) / gsolve->propensity_enz[0][i];
		      if (time < minimum_time)
			{
			  minimum_time = time;
			  r_or_e_type = 2;	/* enz chosen */
			  r_or_e_index = i;
			r_or_e_direction = 1;
			}
		    } else
		      time = DBL_MAX;

		    if (gsolve->propensity_enz[1][i]) {
		      time = -1 * log (sprng ()) / gsolve->propensity_enz[1][i];
		      if (time < minimum_time)
			{
			  minimum_time = time;
			  r_or_e_type = 2;	/* enz chosen */
			  r_or_e_index = i;
			  r_or_e_direction = 2;
			}
		    } else
		      time = DBL_MAX;

		    if (gsolve->propensity_enz[2][i]) {
		      time = -1 * log (sprng ()) / gsolve->propensity_enz[2][i];
		      if (time < minimum_time)
			{
			  minimum_time = time;
			  r_or_e_type = 2;	/* enz chosen */
			  r_or_e_index = i;
			  r_or_e_direction = 3;
			}
		    } else
		      time = DBL_MAX;
		  }
		
		if (r_or_e_type == 1)
		  {
		    reac_chosen =
		      (struct reac_type *) gsolve->reac_list->
		      element[r_or_e_index];
		    reac_chosen_direction = r_or_e_direction;
		  }
		else if (r_or_e_type == 2)
		  {
		    enz_chosen =
		      (struct enz_type *) gsolve->enz_list->element[r_or_e_index];
		    enz_chosen_direction = r_or_e_direction;
		  }
		
		transition_time = minimum_time;
	      }
	    
	    /* 	gsolve->curr_time += transition_time; */
	    
	    /*
	     * update relevant (number of) molecules 
	     */
	    if (reac_chosen)
	      {
		data = reac_chosen->msgdata;
		se = reac_chosen->se;
		switch (reac_chosen_direction)
		  {
		  case 1:		/* forward */
		    for (j = 0; j < reac_chosen->msgcount[0]; j++, data++, se++)
		      if (!(**se & POOL_BUFFER))
			**data -= 1;	/* TYPE = SUBSTRATE */
		    for (j = 0; j < reac_chosen->msgcount[1]; j++, data++, se++)
		      if (!(**se & POOL_BUFFER))
			**data += 1;	/* TYPE = PRODUCT */
		    break;
		  case -1:		/* backward */
		    for (j = 0; j < reac_chosen->msgcount[0]; j++, data++, se++)
		      if (!(**se & POOL_BUFFER))
			**data += 1;	/* TYPE = SUBSTRATE */
		    for (j = 0; j < reac_chosen->msgcount[1]; j++, data++, se++)
		      if (!(**se & POOL_BUFFER))
			**data -= 1;	/* TYPE = PRODUCT */
		    break;
		  default:
		    break;
		  }
	      }
	    if (enz_chosen)
	      {
		data = enz_chosen->msgdata;
		se = enz_chosen->se;
		dt = Clockrate (enz_chosen);
		switch (enz_chosen_direction)
		  {
		  case 1:		/* E + S -----> E.S */
		    enz_chosen->nComplex++;
		    if (enz_chosen->usecomplex) {
		      data += enz_chosen->msgcount[0];
		      se += enz_chosen->msgcount[0];
		    }
		    else
		      for (j = 0; j < enz_chosen->msgcount[0]; j++, data++, se++)
			if (!(**se & POOL_BUFFER))
			  **data -= 1;
		    for (j = 0; j < enz_chosen->msgcount[1]; j++, data++, se++)
		      if (!(**se & POOL_BUFFER))
			**data -= 1;
		    enz_chosen->pA = 0;
		    break;
		    
		  case 2:		/* E.S -----> E + S */
		    enz_chosen->nComplex--;
		    if (enz_chosen->usecomplex) {
		      data += enz_chosen->msgcount[0];
		      se += enz_chosen->msgcount[0];
		    }
		    else
		      for (j = 0; j < enz_chosen->msgcount[0]; j++, data++, se++)
			if (!(**se & POOL_BUFFER))
			  **data += 1;
		    for (j = 0; j < enz_chosen->msgcount[1]; j++, data++, se++)
		      if (!(**se & POOL_BUFFER))
			**data += 1;
		    enz_chosen->pA = 0;
		    break;
		    
		  case 3:		/* E.S -----> E + P */
		    enz_chosen->nComplex--;
		    if (!enz_chosen->usecomplex)
		      for (j = 0; j < enz_chosen->msgcount[0]; j++, data++, se++)
			if (!(**se & POOL_BUFFER))
			  **data += 1;
		    enz_chosen->pA = 1 / dt;
		    break;
		    
		  default:
		    break;
		  }
		
		if (enz_chosen->nComplex < 0)
		  enz_chosen->nComplex = 0;
		if (enz_chosen->usecomplex)
		  {
		    data = enz_chosen->msgdata;
		    for (j = 0; j < enz_chosen->msgcount[0]; j++, data++)
		      orig_nenz = **data;
		    if (enz_chosen->nComplex > orig_nenz)
		      enz_chosen->nComplex = orig_nenz;
		  }
	      }	

	    /* If we are going to shoot by end time, skip out now.
	     * update step and time counts appropriately */
	    if (*simulation_time + transition_time > *start_time + *duration)
	      {
		*current_step = 
		  ((*start_time + *duration) / ClockValue (0)) - 1;
		*stepcount = *current_step; /* aaaaaaargh! */
		*nsteps = 1;
		*simulation_time = (*start_time + *duration) - ClockValue (0);
		
		gsolvePoke (&nsteps, &stepcount, &current_step,
			    &simulation_time);
		goto exitprocess;
	      }
	    
	    /* making appropriate deductions for sim_step's normal
	     *  updates of those variables */
	    
	    *simulation_time += transition_time - ClockValue (0);
	    transition_steps = transition_time / ClockValue (0);
	    /* strange but true */
	    *stepcount += transition_steps - 1;
	    *current_step += transition_steps - 1;
	    *nsteps -= transition_steps - 1;
	  }
	  else { /* gsolve->method == 3; Gibson-Bruck's modification to
		    Gillespie's First Reaction Method */
	    
	    if (gsolve->enz_list_populated)
	      for (i = 0; i < gsolve->enz_list->nelements; i++) {
		enz = (struct enz_type *) gsolve->enz_list->element[i];
		enz->pA = 0;
	      }
	    
	    if (!(gsolve->propensity_total)) {
	      *current_step = ((*start_time + *duration) / ClockValue(0)) - 1;
	      *stepcount = *current_step;
	      *nsteps = 1;
	      *simulation_time = (*start_time + *duration) - ClockValue(0);
          
	      gsolvePoke(&nsteps, &stepcount, &current_step, &simulation_time);
	      goto exitprocess;
	    }

	    trans_chosen = pq_nodes[0].transition_index;
	    transition_time = pq_nodes[0].time;
	    
	    if (trans_chosen < 1000) {     /* reac chosen */
	      reac_chosen = (struct reac_type *) gsolve->reac_list->element[trans_chosen / 2];
	      reac_chosen_direction = trans_chosen % 2;

	      data = reac_chosen->msgdata;
	      se = reac_chosen->se;
	      switch (reac_chosen_direction) {
	      case 0:   /* forward */
		for (j = 0; j < reac_chosen->msgcount[0]; j++, data++, se++)
		  if (!(**se & POOL_BUFFER))
		    **data -= 1;
		for (j = 0; j < reac_chosen->msgcount[1]; j++, data++, se++)
		  if (!(**se & POOL_BUFFER))
		    **data += 1;
		break;
	      case 1:   /* backward */
		for (j = 0; j < reac_chosen->msgcount[0]; j++, data++, se++)
		  if (!(**se & POOL_BUFFER))
		    **data += 1;
		for (j = 0; j < reac_chosen->msgcount[1]; j++, data++, se++)
		   if (!(**se & POOL_BUFFER))
		    **data -= 1;
		break;
	      default:
		break;
	      }
	    }
	    else {                         /* enz chosen */
	      enz_chosen = (struct enz_type *) gsolve->enz_list->element[(trans_chosen - 1000) / 3];
	      enz_chosen_direction = (trans_chosen - 1000) % 3;
	      
	      data = enz_chosen->msgdata;
	      se = enz_chosen->se;
	      dt = Clockrate (enz_chosen);
	      switch (enz_chosen_direction) {
	      case 0:   /* E + S -----> E.S */
                enz_chosen->nComplex++;
                if (enz_chosen->usecomplex) {
                  data += enz_chosen->msgcount[0];
		   se += enz_chosen->msgcount[0];
		}
                else
                  for (j = 0; j < enz_chosen->msgcount[0]; j++, data++, se++)
		    if (!(**se & POOL_BUFFER))
		      **data -= 1;
                for (j = 0; j < enz_chosen->msgcount[1]; j++, data++, se++)
		  if (!(**se & POOL_BUFFER))
		    **data -= 1;
                enz_chosen->pA = 0;
		break;
	      case 1:   /* E.S -----> E + S */
                enz_chosen->nComplex--;
                if (enz_chosen->usecomplex) {
                  data += enz_chosen->msgcount[0];
		  se += enz_chosen->msgcount[0];
		}
                else
                  for (j = 0; j < enz_chosen->msgcount[0]; j++, data++, se++)
		    		    if (!(**se & POOL_BUFFER))
		      **data += 1;
                for (j = 0; j < enz_chosen->msgcount[1]; j++, data++, se++)
		  if (!(**se & POOL_BUFFER))
		    **data += 1;
                enz_chosen->pA = 0;
		break;
	      case 2:   /* E.S -----> E + P */
                enz_chosen->nComplex--;
                if (!enz_chosen->usecomplex)
                  for (j = 0; j < enz_chosen->msgcount[0]; j++, data++, se++)
		    if (!(**se & POOL_BUFFER))
		      **data += 1;
                enz_chosen->pA = 1 / dt;
		break;
	      default:
		break;
	      }
	      if (enz_chosen->nComplex < 0)
		enz_chosen->nComplex = 0;
	      if (enz_chosen->usecomplex)
		{
		  data = enz_chosen->msgdata;
		  for (j = 0; j < enz_chosen->msgcount[0]; j++, data++)
		    orig_nenz = **data;
		  if (enz_chosen->nComplex > orig_nenz)
		    enz_chosen->nComplex = orig_nenz;
		}
	    }

	    for (i = 0; i < dep_graph_adj[trans_chosen][1900]; i++) {
	      k = dep_graph_adj[trans_chosen][i];
	      propensity_trans_old = propensity_trans[k];
	      if (k < 1000) {
		gsolve->propensity_total -= propensity_trans[k];
		reac = (struct reac_type *) gsolve->reac_list->element[k / 2];
		data = reac->msgdata;
		
		if ((k % 2) == 0) {
		  propensity_trans[k] = reac->kf;
		  for (j = 0; j < reac->msgcount[0]; j++, data++)
		    propensity_trans[k] *= **data;
		}
		else {
		  data += reac->msgcount[0];
		  propensity_trans[k] = reac->kb;
		  for (j = 0; j < reac->msgcount[1]; j++, data++)
		    propensity_trans[k] *= **data;
		}
		gsolve->propensity_total += propensity_trans[k];
	      }
	      else {
		gsolve->propensity_total -= propensity_trans[k];
		enz = (struct enz_type *) gsolve->enz_list->element[(k - 1000) / 3];
		data = enz->msgdata;
		
		if (((k - 1000) % 3) == 0) {
		  for (j = 0; j < enz->msgcount[0]; j++, data++)
		    propensity_trans[k] = **data;
		  if (enz->usecomplex)
		    propensity_trans[k] -= enz->nComplex;
		  if (propensity_trans[k] < 1)
		    propensity_trans[k] = 0;
		  for (j = 0; j < enz->msgcount[1]; j++, data++)
		    propensity_trans[k] *= **data;
		  propensity_trans[k] *= enz->k1;
		}
		else if (((k - 1000) % 3) == 1) {
		  propensity_trans[k] = enz->k2 * enz->nComplex;
		  if (enz->nComplex < 1)
		    propensity_trans[k] = 0;
		}
		else {
		  propensity_trans[k] = enz->k3 * enz->nComplex;
		  if (enz->nComplex < 1)
		    propensity_trans[k] = 0;
		}
		
		gsolve->propensity_total += propensity_trans[k];
	      }
	      if (k != trans_chosen) {
		if (propensity_trans_old) {
		  if (propensity_trans[k]) {
		    pq_update(transition_pq_index[k],  
			      ((propensity_trans_old / propensity_trans[k])  
			       * (pq_nodes[transition_pq_index[k]].time - transition_time))  
			      + transition_time);
		  } else {
		    propensity_trans_zero[k] = propensity_trans_old;
		    time_zero[k] = transition_time;
		    trans_time_zero[k] = pq_nodes[transition_pq_index[k]].time;
		    pq_update(transition_pq_index[k], DBL_MAX);
		  }
		} else {   // propensity_trans_old == 0
		  if (propensity_trans[k]) {
		    if (propensity_trans_zero[k])
		      pq_update(transition_pq_index[k],
				((propensity_trans_zero[k] / propensity_trans[k]) 
				 * (trans_time_zero[k] - time_zero[k])) 
				+ transition_time);
		    else
		      pq_update(transition_pq_index[k],
				-1 * log (sprng ()) / propensity_trans[k]
				+ transition_time);
		  } else {
		    // Do nothing. propensity should already be DBL_MAX
		  }
		}
	      }
	      else {
		if (propensity_trans[k])
		  pq_update(transition_pq_index[k],
			    -1 * log (sprng ()) / propensity_trans[k]
			    + transition_time);
		else {
		  pq_update(transition_pq_index[k], DBL_MAX);
		  propensity_trans_zero[k] = 0;
		}
	      }
	    }
	    
#if 0
	    if (!(gsolve->propensity_total)) {
	      *current_step = ((*start_time + *duration) / ClockValue(0)) - 1;
	      *stepcount = *current_step;
	      *nsteps = 1;
	      *simulation_time = (*start_time + *duration) - ClockValue(0);
          
	      gsolvePoke(&nsteps, &stepcount, &current_step, &simulation_time);
	      goto exitprocess;
	    }
#endif

	    if (transition_time > *start_time + *duration) {
	      *current_step = ((*start_time + *duration) / ClockValue(0)) - 1;
	      *stepcount = *current_step;
	      *nsteps = 1;
	      *simulation_time = (*start_time + *duration) - ClockValue(0);
	      
	      gsolvePoke (&nsteps, &stepcount, &current_step,
			  &simulation_time);
	      goto exitprocess;
	    }
	    transition_steps = (transition_time - *simulation_time) / ClockValue(0);
	    *stepcount += transition_steps - 1;
	    *current_step += transition_steps - 1;
	    *nsteps -= transition_steps - 1;
	    *simulation_time = transition_time - ClockValue(0);
	  }
	  gsolvePoke (&nsteps, &stepcount, &current_step, &simulation_time);
	}
      /*
       * GsolveBaseProcess (gsolve); 
       */
    exitprocess:
      break;
    case RESET:
      gsolve_do_reset (gsolve);
      break;
    case CREATE:
      /* gsolve_do_reset (gsolve); *//* Don't do this! */
      break;
    case SET:
      return (0);
      break;
    }
}

#ifdef __STDC__
static int
check_msg_type (struct gsolve_type *gsolve, MsgIn * msg)
{
#else
static int
check_msg_type (gsolve, msg)
     struct gsolve_type *gsolve;
     MsgIn *msg;
{
#endif
  if (MSGSLOT (msg)[0].func != DoubleMessageData)
    {
      Warning ();
      /*
       * Complain 
       */
      printf ("Message source for REAC msg to %s is not a double\n",
	      Pathname (gsolve));
      printf ("Reverting to slow old messaging\n");
      gsolve->oldmsgmode = 1;
    }
}

static int
gsolve_do_reset (struct gsolve_type *gsolve)
{
  MsgIn *msg;
  MsgOut *mout;
  
  int i, j, j2, k, t;
  double **data;
  struct reac_type *reac;
  struct enz_type *enz;
  struct pool_type *pool, *sumtotal_src;
  
  int fwd_successors[200]; /* assuming a maximum of 200 successors of a pool */
  int fwd_reac_successors = 0;
  int bwd_successors[200];
  int bwd_reac_successors = 0;
  
  Msg *msgout, *msgin, *msgin2;
  unsigned int nmsgout, nmsgin, nmsgin2;
  MsgList *ml, *ml2;
  GenesisObject *object, *object2;

  int *nsteps;
  int *stepcount;
  int *current_step;
  float *start_time;
  float *duration;
  double *simulation_time;

  int dummy_int = 0;
  int **se;

  gsolve->propensity_total = 0.0;
  
  gsolve->reac_list = WildcardGetElement ("##[TYPE=kreac]", 0);
  if ((!gsolve->reac_list) || (gsolve->reac_list->nelements == 0))
    gsolve->reac_list_populated = 0;
  else
    {
      gsolve->reac_list_populated = 1;
      ResetElements (gsolve->reac_list);
    }
  
  gsolve->enz_list = WildcardGetElement ("##[TYPE=kenz]", 0);
  if ((!gsolve->enz_list) || (gsolve->enz_list->nelements == 0))
    gsolve->enz_list_populated = 0;
  else
    {
      gsolve->enz_list_populated = 1;
      ResetElements (gsolve->enz_list);
    }
  
  gsolve->pool_list = WildcardGetElement ("##[TYPE=kpool]",0);
  if ((!gsolve->pool_list) || (gsolve->pool_list->nelements == 0))
    gsolve->pool_list_populated = 0;
  else
    gsolve->pool_list_populated = 1;

  /* flooring all pool and enzyme complex 'n' fields once */
  for (i = 0; i < gsolve->pool_list->nelements; i++)
    {
      pool = (struct pool_type *) gsolve->pool_list->element[i];
      pool->n = floor(pool->nInit);
    }
  
  for (i = 0; i < gsolve->enz_list->nelements; i++)
    {
      enz = (struct enz_type *) gsolve->enz_list->element[i];
      enz->nComplex = floor(enz->nComplex);
    }
    
    
  /* Populating dep_graph_index */
  for (i = 0; i < gsolve->reac_list->nelements; i++)
    {
      reac = (struct reac_type *) gsolve->reac_list->element[i];
      reac->A = reac->B = 0;
      reac->dep_graph_index = i*2; /* Two transitions per reac */
    }
  
  for (i = 0; i < gsolve->enz_list->nelements; i++)
    {
      enz = (struct enz_type *) gsolve->enz_list->element[i];
      enz->sA = enz->eA = enz->B = 0;
      enz->dep_graph_index = 1000 + i*3; /* Three transitions per enz; enzymes begin at 1000 */
    }
  
  if (gsolve->method == 3) { /* Gibson-Bruck */
    /* Build dependency graph */

    /* normal reacs, via pool_list */
    /* forward: pool incoming REACs are linked to pool outgoing SUBSTRATEs */
    /* backward: pool incoming REACs are linked to pool outgoing PRODUCTs */
    
    /* enzymes: (internal), via enz_list */
         
    /* TODO Profile use of adj matrix per timestep.. switch to adj list if too slow
	* 		This _is_ per PROCESS */
    
    /* TODO Pool to pool CONSERVE/SUMTOTAL handling - nothing special needed:? */

    gsolvePeek(&nsteps, &stepcount, &current_step, &start_time, &duration, 
               &simulation_time); 
    
    for (i = 0; i < 1900; i++) {
      propensity_trans[i] = 0;
      propensity_trans_zero[i] = 0;
      time_zero[i] = *simulation_time;
      trans_time_zero[i] = DBL_MAX;
      transition_pq_index[i] = 1900;  /* never an actual point in pq */
      pq_nodes[i].time = DBL_MAX;
      pq_nodes[i].transition_index = 1900; /* never an actual transition */
    }

    for (i = 0; i < 1900; i++)
      for (j = 0; j < 1901; j++)
	dep_graph_adj[i][j] = 0;

    pq_size = 0;
    
    for (i = 0; i < gsolve->pool_list->nelements; i++) {
      pool = (struct pool_type *) gsolve->pool_list->element[i];
      msgout = pool->msgout;
      nmsgout = pool->nmsgout;
      // TODO is pool->msgout TRUE?
      for (j = 0; j < nmsgout; j++) {
	if (VISIBLE(msgout->dst)) {
	  ml = GetMsgListByType(msgout->dst->object, msgout->type);
	  // TODO optimize or replace slow string comparison
	  if ((ml) && (strcmp(ml->name, "SUBSTRATE") == 0)) {
	    if (ElementIsA(msgout->dst, "reac")) {
	      reac = (struct reac_type *) msgout->dst;
	      fwd_successors[fwd_reac_successors] = reac->dep_graph_index;
	      fwd_reac_successors++;
	    } else if (ElementIsA(msgout->dst, "enz")) {
	      enz = (struct enz_type *) msgout->dst;
	      fwd_successors[fwd_reac_successors] = enz->dep_graph_index;
	      fwd_reac_successors++;
	    }
	  }
	  else if ((ml) && (strcmp(ml->name, "ENZYME") == 0)) {
	    enz = (struct enz_type *) msgout->dst;
	    fwd_successors[fwd_reac_successors] = enz->dep_graph_index;
	    fwd_reac_successors++;
	  }
	  else if ((ml) && (strcmp(ml->name, "PRODUCT") == 0) && (ElementIsA(msgout->dst, "reac"))) {
	    reac = (struct reac_type *) msgout->dst;
	    bwd_successors[bwd_reac_successors] = reac->dep_graph_index + 1;
	    bwd_reac_successors++;
	  }
	}
	msgout = MSGOUTNEXT(msgout);
      }
      msgin = pool->msgin;
      object = pool->object;
      nmsgin = pool->nmsgin;
      // TODO is pool->msgin TRUE?
      for (j = 0; j < nmsgin; j++) {
	if (VISIBLE(msgin->src)) {
	  ml = GetMsgListByType(object, msgin->type);
	  if ((ml) && (strcmp(ml->name, "REAC") == 0)) {
	    if (ElementIsA(msgin->src, "reac")) {
	      reac = (struct reac_type *) msgin->src;
	      t = reac->dep_graph_index;
	      for (k = 0; k < fwd_reac_successors; k++) {
		add_dep_list(t, fwd_successors[k]);
		add_dep_list(t+1, fwd_successors[k]);
	      }
	      for (k = 0; k < bwd_reac_successors; k++) {
		add_dep_list(t, bwd_successors[k]);
		add_dep_list(t+1, bwd_successors[k]);
	      }
	    } else if (ElementIsA(msgin->src, "enz")) {
	      enz = (struct enz_type *) msgin->src;
	      t = enz->dep_graph_index;
	      // TODO *_reac_* is a bad choice of name; change this appropriately
	      for (k = 0; k < fwd_reac_successors; k++) {
		add_dep_list(t, fwd_successors[k]);
		add_dep_list(t+1, fwd_successors[k]);
	      }
	      for (k = 0; k < bwd_reac_successors; k++) {
		add_dep_list(t, bwd_successors[k]);
		add_dep_list(t+1, bwd_successors[k]);
	      }
	    }
	  }
	  else if ((ml) && (strcmp(ml->name, "MM_PRD") == 0)) {
	    enz = (struct enz_type *) msgin->src;
	    t = enz->dep_graph_index;
	    for (k = 0; k < fwd_reac_successors; k++)
	      add_dep_list(t+2, fwd_successors[k]);
	    for (k = 0; k < bwd_reac_successors; k++)
	      add_dep_list(t+2, bwd_successors[k]);
	  }
	}
	msgin = MSGINNEXT(msgin);
      }

      msgin = pool->msgin;
      object = pool->object;
      nmsgin = pool->nmsgin;
      for (j = 0; j < nmsgin; j++) {
	if (VISIBLE(msgin->src)) {
	  ml = GetMsgListByType(object, msgin->type);
	  if ((ml) && (strcmp(ml->name, "SUMTOTAL") == 0)) {
	    // assuming only POOLs send out SUMTOTALs
	    sumtotal_src = (struct pool_type *) msgin->src;
	    
	    msgin2 = sumtotal_src->msgin;
	    object2 = sumtotal_src->object;
	    nmsgin2 = sumtotal_src->nmsgin;

	    for (j2 = 0; j2 < nmsgin2; j2++) {
	      if (VISIBLE(msgin2->src)) {
		ml2 = GetMsgListByType(object2, msgin2->type);
		if ((ml2) && (strcmp(ml2->name, "REAC") == 0)) {
		  if (ElementIsA(msgin2->src, "reac")) {
		    reac = (struct reac_type *) msgin2->src;
		    t = reac->dep_graph_index;
		    for (k = 0; k < fwd_reac_successors; k++) {
		      add_dep_list(t, fwd_successors[k]);
		      add_dep_list(t+1, fwd_successors[k]);
		    }
		    for (k = 0; k < bwd_reac_successors; k++) {
		      add_dep_list(t, bwd_successors[k]);
		      add_dep_list(t+1, bwd_successors[k]);
		    }
		  } else if (ElementIsA(msgin2->src, "enz")) {
		    enz = (struct enz_type *) msgin2->src;
		    t = enz->dep_graph_index;
		    for (k = 0; k < fwd_reac_successors; k++) {
		      add_dep_list(t, fwd_successors[k]);
		      add_dep_list(t+1, fwd_successors[k]);
		    }
		    for (k = 0; k < bwd_reac_successors; k++) {
		      add_dep_list(t, bwd_successors[k]);
		      add_dep_list(t+1, bwd_successors[k]);
		    }
		  }
		} else if ((ml2) && (strcmp(ml2->name, "MM_PRD") == 0)) {
		  enz = (struct enz_type *) msgin2->src;
		  t = enz->dep_graph_index;
		  for (k = 0; k < fwd_reac_successors; k++)
		    add_dep_list(t+2, fwd_successors[k]);
		  for (k = 0; k < bwd_reac_successors; k++)
		    add_dep_list(t+2, bwd_successors[k]);
		}
	      }
	      msgin2 = MSGINNEXT(msgin2);
	    }
	  }
	}
	msgin = MSGINNEXT(msgin);
      }
      
      fwd_reac_successors = 0;
      bwd_reac_successors = 0;
    }

    if (gsolve->enz_list_populated) {
      for (i = 0; i < gsolve->enz_list->nelements; i++) {
	enz = (struct enz_type *) gsolve->enz_list->element[i];
	t = enz->dep_graph_index;
	for (j = 0 ; j < 3; j++)
	  for (k = 0; k < 3; k++)
	    add_dep_list(t+j, t+k);
      }
    }

    if (gsolve->reac_list_populated)
      for (i = 0; i < gsolve->reac_list->nelements; i++) {
	reac = (struct reac_type *) gsolve->reac_list->element[i];
	data = reac->msgdata;

	propensity_trans[i*2] = reac->kf;
	for (j = 0; j < reac->msgcount[0]; j++, data++)
	  propensity_trans[i*2] *= **data;
	
	propensity_trans[1+i*2] = reac->kb;
	for (j = 0; j < reac->msgcount[1]; j++, data++)
	  propensity_trans[1+i*2] *= **data;
	
	gsolve->propensity_total +=
	  propensity_trans[i*2] + propensity_trans[1+i*2];
      }
    
    if (gsolve->enz_list_populated)
      for (i = 0; i < gsolve->enz_list->nelements; i++) {
	enz = (struct enz_type *) gsolve->enz_list->element[i];
	data = enz->msgdata;
	
	for (j = 0; j < enz->msgcount[0]; j++, data++)
	  propensity_trans[1000 + i*3] = **data;
	if (enz->usecomplex)
	  propensity_trans[1000 + i*3] -= enz->nComplex;
	if (propensity_trans[1000 + i*3] < 1)
	  propensity_trans[1000 + i*3] = 0;
	for (j = 0; j < enz->msgcount[1]; j++, data++)
	  propensity_trans[1000 + i*3] *= **data;
	propensity_trans[1000 + i*3] *= enz->k1;
	
	propensity_trans[1001 + i*3] = enz->k2 * enz->nComplex;
	if (enz->nComplex < 1)
	  propensity_trans[1001 + i*3] = 0;
	
	propensity_trans[1002 + i*3] = enz->k3 * enz->nComplex;
	if (enz->nComplex < 1)
	  propensity_trans[1002 + i*3] = 0;
	
	gsolve->propensity_total +=
	  propensity_trans[1000 + i*3] + propensity_trans[1001 + i*3]
	  + propensity_trans[1002 + i*3];
	
      }
    /* Populate priority queue storing (absolute) transition times */
    
    for (i = 0; i < gsolve->reac_list->nelements; i++) {
      if (propensity_trans[i*2])
	pq_insert((-1 * log (sprng ()) / propensity_trans[i*2]) + *simulation_time, i*2);
      else
	pq_insert(DBL_MAX, i*2);
      if (propensity_trans[1+i*2])
	pq_insert((-1 * log (sprng ()) / propensity_trans[1+i*2]) + *simulation_time, 1+i*2);
      else
	pq_insert(DBL_MAX, 1+i*2);
    }
    for (i = 0; i < gsolve->enz_list->nelements; i++) {
      if (propensity_trans[1000+i*3])
	pq_insert((-1 * log (sprng ()) / propensity_trans[1000+i*3]) + *simulation_time, 1000+i*3);
      else
	pq_insert(DBL_MAX, 1000+i*3);
      if (propensity_trans[1001+i*3])
	pq_insert((-1 * log (sprng ()) / propensity_trans[1001+i*3]) + *simulation_time, 1001+i*3);
      else
	pq_insert(DBL_MAX, 1001+i*3);
      if (propensity_trans[1002+i*3])
	pq_insert((-1 * log (sprng ()) / propensity_trans[1002+i*3]) + *simulation_time, 1002+i*3);
      else
	pq_insert(DBL_MAX, 1002+i*3);
    } 
  }
  /* Common for all Gillespie solvers */
  /* Building a table per reac and per enz of slave_enable flags of 
     leading pools */
  for (i = 0; i < gsolve->reac_list->nelements; i++) {
    reac = (struct reac_type *) gsolve->reac_list->element[i];
    se = reac->se;
    MSGLOOP(reac,msg) {
    case 0:  /* TYPE = SUBSTRATE */
      if (ElementIsA(msg->src, "pool"))
	*se++ = &((struct pool_type *) (msg->src))->slave_enable;
      else
	*se++ = &dummy_int;
      break;
    case 1:  /* TYPE = PRODUCT */
      if (ElementIsA(msg->src, "pool"))
	*se++ = &((struct pool_type *) (msg->src))->slave_enable;
      else
	*se++ = &dummy_int;
      break;
    }
  }
  
  for (i = 0; i < gsolve->enz_list->nelements; i++) {
    enz = (struct enz_type *) gsolve->enz_list->element[i];
    se = enz->se;
    MSGLOOP(enz,msg) {
    case 0:  /* TYPE = SUBSTRATE */
      if (ElementIsA(msg->src, "pool"))
	*se++ = &((struct pool_type *) (msg->src))->slave_enable;
      else
	*se++ = &dummy_int;
      break;
    case 1:  /* TYPE = PRODUCT */
      if (ElementIsA(msg->src, "pool"))
	*se++ = &((struct pool_type *) (msg->src))->slave_enable;
      else
	*se++ = &dummy_int;
      break;
    }
    MSGOUTLOOP(enz,msg) {
      if (ElementIsA(msg->dst, "pool") && msg->type == 1)
	*se++ = &((struct pool_type *) (msg->dst))->slave_enable;
      else
	*se++ = &dummy_int;
    }
  }
}

/*
 * This function is declared as an extern void() 
 */
/*
 * void copyleft_kin () { printf ("The kinetics library is copylefted
 * under the LGPL, see kinetics/COPYRIGHT.\n\n"); }
 */

// Priority Queue housekeeping
// INSERT
void pq_insert(double t, int t_i) {
  int m = pq_size;

  pq_nodes[m].time = t;
  pq_nodes[m].transition_index = t_i;
  transition_pq_index[t_i] = m;

  while (m != 0) {    
    if (pq_nodes[m].time < pq_nodes[m/2].time) {
      pq_swap(m, m/2);
      m = m/2;
    }
    else
      break;
  }

  pq_size++;
  return;
}

// SWAP (and update index)
void pq_swap(int a, int b) {
  pq_node temp;

  temp = pq_nodes[a];
  pq_nodes[a] = pq_nodes[b];
  pq_nodes[b] = temp;

  transition_pq_index[pq_nodes[a].transition_index] = a;
  transition_pq_index[pq_nodes[b].transition_index] = b;

  return;
}


// UPDATE
void pq_update(int n, double new_t) {
  pq_nodes[n].time = new_t;
  pq_update_aux(n);

  return;
}

void pq_update_aux(int n) {
  double min_val_children = 0;
  int min_child = 0;

  if (pq_nodes[n].time < pq_nodes[n/2].time) {
    pq_swap(n, n/2);
    pq_update_aux(n/2);
  }
  else if ((2*n) < 1900) {
    if (pq_nodes[2*n].time < pq_nodes[2*n+1].time) {
      min_val_children = pq_nodes[2*n].time;
      min_child = 2*n;
    } else {
      min_val_children = pq_nodes[2*n+1].time;
      min_child = 2*n+1;
    }
    if (pq_nodes[n].time > min_val_children) {
      pq_swap(n, min_child);
      pq_update_aux(min_child);
    } 
  }

  return;
}

int add_dep_list(int trans /* ition */, int dep /* endent */ ) {
  int i, already_in = 0;

  for (i = 0; i < dep_graph_adj[trans][1900]; i++) {
    if (dep_graph_adj[trans][i] == dep)
      already_in = 1;
  }

  if (already_in)
    return -1;
  else {
    dep_graph_adj[trans][dep_graph_adj[trans][1900]++] = dep;
    return 0;
  }
}

// TODO: copy by reference instead of bitwise (OPTI)

#undef POOL_BUFFER
