/**********************************************************************
** This program is part of the kinetics library and is
**           copyright (C) 1995 Upinder S. Bhalla.
** It is made available under the terms of the
**           GNU Library General Public License. 
** See the file COPYRIGHT for the full notice.
**********************************************************************/
static char rcsid[] = "$Id: reac.c,v 1.3 1998/07/21 19:49:11 dhb Exp $";
 
/*
** $Log: reac.c,v $
** Revision 1.3  1998/07/21 19:49:11  dhb
** Fixed ANSI C isms.
**
 * Revision 1.2  1998/07/15 06:45:33  dhb
 * Upi update
 *
 * Revision 1.1  1997/07/24 17:49:40  dhb
 * Initial revision
 *
 * Revision 1.2  1994/08/05  19:24:43  bhalla
 * Minor changes due to conversion to using n rather than conc
 *
 * Revision 1.1  1994/06/13  22:55:39  bhalla
 * Initial revision
 * */

#include "kin_ext.h"
#ifdef CHECK
#undef CHECK
#define SIMPLE_SPRNG
#include "sprng.h"
#endif


/*
** A reaction handling element. does the basic forward/backward
** rate const management, must be used in conjunction with pools.
** Can handle any number of substrate or product components, but
** the reaction is only a single stage
*/

#ifdef __STDC__
static int do_oldmsgmode(struct reac_type *reac);
static int check_msg_type(struct reac_type *reac, MsgIn *msg);
#else
static int do_oldmsgmode();
static int check_msg_type();
#endif

ReacContProcess(struct reac_type* reac) {
	double substrate_prd = reac->kf;
	double product_prd = reac->kb;
	double **data = reac->msgdata;
	int j;

	/* TYPE = SUBSTRATE. */
	for (j = 0; j < reac->msgcount[0]; j++, data++)
		substrate_prd *= **data;
	/* TYPE = PRODUCT. */
	for (j = 0; j < reac->msgcount[1]; j++, data++)
		product_prd *= **data;

	reac->A = product_prd;
	reac->B = substrate_prd;
}

ReacStochProcess(struct reac_type* reac) {
	double dt = Clockrate(reac);
	double substrate_prd = reac->kf * dt;
	double product_prd = reac->kb * dt;
	double **data = reac->msgdata;
	int j;
	double dn;
	double minsub = KIN_STOCH_UPPER_MOL_LIMIT;
	double minprd = KIN_STOCH_UPPER_MOL_LIMIT;
	double minmol;
	double n;

	/* TYPE = SUBSTRATE. */
	for (j = 0; j < reac->msgcount[0]; j++, data++) {
			n = **data;
			substrate_prd *= n;
			if (minsub > n) minsub = n;
	}

	/* TYPE = PRODUCT. */
	for (j = 0; j < reac->msgcount[1]; j++, data++) {
		n = **data;
		product_prd *= n;
		if (minprd > n) minprd = n;
	}


	minmol = (minsub < minprd) ? minsub : minprd;

	if (reac->stoch_flag > 1) {
			if (minmol < KIN_STOCH_LOWER_MOL_LIMIT)
				reac->stoch_flag = 1;
			else if ( substrate_prd < KIN_STOCH_LOWER_PROB_LIMIT &&
				product_prd < KIN_STOCH_LOWER_PROB_LIMIT )
				reac->stoch_flag--;
	} else if ( minmol >= KIN_STOCH_UPPER_MOL_LIMIT &&
				(substrate_prd > KIN_STOCH_UPPER_PROB_LIMIT ||
				product_prd > KIN_STOCH_UPPER_PROB_LIMIT) ) {
				reac->stoch_flag = KIN_STOCH_COUNTDOWN;
	}
	if (minmol < KIN_STOCH_LOWER_MOL_LIMIT && 
				(substrate_prd > KIN_STOCH_UPPER_PROB_LIMIT ||
				product_prd > KIN_STOCH_UPPER_PROB_LIMIT) ) {
			reac->propensity_problem++;
	}

	if (reac->stoch_flag == 1) {
#if 0
		substrate_prd = reac->kf * dt;
		product_prd = reac->kb * dt;
		/* TYPE = SUBSTRATE. */
		data = reac->msgdata;
		for (j = 0; j < reac->msgcount[0]; j++, data++)
			substrate_prd *= (int)(**data + KIN_STOCH_EPSILON);

		/* TYPE = PRODUCT. */
		for (j = 0; j < reac->msgcount[1]; j++, data++)
			product_prd *= (int)(**data + KIN_STOCH_EPSILON);
#endif
		dn = (sprng() < product_prd) - (sprng() < substrate_prd);
		/* Check for cases where we try to remove a molecule that
		 * doesn't exist */
		if ((minsub < 1 && dn < -0.1) || (minprd < 1 && dn > 0.1))
				dn = 0;

		reac->A = reac->B = 0;

		data = reac->msgdata;
		/* TYPE = SUBSTRATE. */
		for (j = 0; j < reac->msgcount[0]; j++, data++)
			**data += dn;

		/* TYPE = PRODUCT. */
		for (j = 0; j < reac->msgcount[1]; j++, data++)
			**data -= dn;
			
			/*
			reac->A = (sprng() < product_prd * dt) / dt;
			reac->B = (sprng() < substrate_prd * dt) / dt;
			*/
	} else {
			reac->A = product_prd/dt;
			reac->B = substrate_prd/dt;
	}
}

ReacStochCheck(struct reac_type* reac) {
	double substrate_prd = reac->kf;
	double product_prd = reac->kb;
	double **data = reac->msgdata;
	int j;
	double dt = Clockrate(reac);

	/* TYPE = SUBSTRATE. */
	for (j = 0; j < reac->msgcount[0]; j++, data++)
		substrate_prd *= **data;

	/* TYPE = PRODUCT. */
	for (j = 0; j < reac->msgcount[1]; j++, data++)
		product_prd *= **data;

	reac->A = product_prd * dt;
	reac->B = substrate_prd * dt;
}

ReacFunc(reac,action)
register struct reac_type *reac;
Action		*action;
{
MsgIn	*msg;
double	substrate_prd;
double	product_prd;
double	dt;
double **data;
int j;

    if(debug > 1){
		ActionHeader("Reaction",reac,action);
    }

    SELECT_ACTION(action){
		case INIT:
			break;
    	case PROCESS:
				if (reac->oldmsgmode)
					return do_oldmsgmode(reac);
				if (reac->stoch_flag)
					ReacStochProcess(reac);
				else
					ReacContProcess(reac);
			break;
    	case RESET: {
				double *tempdata[200];
				data = tempdata;
				reac->msgcount[0] = reac->msgcount[1] = 0;
        		MSGLOOP(reac,msg) {
						case 0: /* TYPE = SUBSTRATE */
							*data++ = (double*)(MSGPTR(msg,0));
							reac->msgcount[0]++;
							break;
						default:
							check_msg_type(reac,msg);
						break;
				}
        		MSGLOOP(reac,msg) {
						case 1: /* TYPE = PRODUCT */
							*data++ = (double*)(MSGPTR(msg,0));
							reac->msgcount[1]++;
							break;
				}
				j = reac->msgcount[0] + reac->msgcount[1];
		        reac->msgdata = alloc_msgdata(j, tempdata);


        		MSGLOOP(reac,msg) {
						case 0:
							break;
						case 1:
							break;
						case 2:		/* TYPE = KF */
									/* msg 0 = kf */
							reac->kf = MSGVALUE(msg,0);
						break;
						case 3:		/* TYPE = KB */
									/* msg 0 = kb */
							reac->kb = MSGVALUE(msg,0);
						break;
        		}
				reac->propensity_problem = 0;
				if (reac->stoch_flag)
					ReacStochCheck(reac);
			}
        	break;
    	case SET :
        	return(0); /* do the normal set */
			break;
		case CREATE:
			reac->oldmsgmode = 0;
			break;
	}
}

#ifdef __STDC__
static int do_oldmsgmode(struct reac_type *reac) {
#else
static int do_oldmsgmode(reac)
struct reac_type *reac;
{
#endif
	MsgIn	*msg;
	double	substrate_prd = reac->kf;
	double	product_prd = reac->kb;
   	MSGLOOP(reac,msg) {
   		case 0:		/* TYPE = SUBSTRATE. */
			/* msg 0 = # of sub. Used to be Co */
			substrate_prd *= MSGVALUE(msg,0);
   		break;
		case 1:		/* TYPE = PRODUCT */
				/* msg 0 = # of prd. Used to be Co */
			product_prd *= MSGVALUE(msg,0);
		break;
		case 2:		/* TYPE = KF */
					/* msg 0 = kf */
			reac->kf = MSGVALUE(msg,0);
		break;
		case 3:		/* TYPE = KB */
					/* msg 0 = kb */
			reac->kb = MSGVALUE(msg,0);
		break;
   	}
	reac->A = product_prd;
	reac->B = substrate_prd;
	return 1;
}

#ifdef __STDC__
static int check_msg_type(struct reac_type *reac, MsgIn *msg) { 
#else
static int check_msg_type(reac, msg)
struct reac_type *reac;
MsgIn *msg;
{ 
#endif
    if (MSGSLOT(msg)[0].func != DoubleMessageData) {
        Warning();
        /* Complain */
        printf("Message source for REAC msg to %s is not a double\n",
        Pathname(reac));
        printf("Reverting to slow old messaging\n");
        reac->oldmsgmode = 1;
    }
}

